#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>
#include "wwork.h"
#include "wvarpi.h"
#include "warpi.h"
#include "types.h"

int setx;
int sety;

U16 startx;
U16 starty;
U8  valerr;
//предельные и средне значения кадра
U8  levmin; //мнимальное
U8  levmax; //маклсимальное
U8  levn;   //среднее
//
pthread_t               hWorkProcess;
pthread_mutex_t         WorkMutex;
int                     WorkWork;
//
U8  wvbuf[640][480];
//координаты начала региона
int strtxr;
int strtyr;
//размер региона
int szxr;
int szyr;
//шаг в регионе
int stepr;
U8  region[640][480];
//
HIST    hist;
HIST    histregion;
//
int     searchx;
int     searchy;
//
HISTSEARCH histsearch;

//запуск рабочего потока
int StartWork(void)
{
    pthread_attr_t 	tattr;
    //параметры для отделяемых потоков
    pthread_attr_init(&tattr);
    pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
    //
    //mutex create
    if(!(pthread_mutex_init(&WorkMutex,NULL) == 0))
        return 3;
    pthread_create(&hWorkProcess, &tattr, wWork_process, NULL);
    WorkWork = 1; //установка флага работы
    //
    return 0;
}

//подсчёт мин. сред. макс.
//для кадра
void wMathExtr(void)
{
    //U8  ln;
    U8  lev;
    int lcntx = 0;
    int lcnty = 0;
    int sumlev = 0;
    levmin = 255;
    levmax = 0;
    while(lcnty < 480)
    {
        lcntx = 0;
        while(lcntx < 640)
        {
            lev = wvbuf[lcntx][lcnty];
            sumlev += lev;
            if(lev < levmin)
                levmin = lev;
            if(lev > levmax)
                levmax = lev;
            lcntx++;
        }
        lcnty++;
    }
    levn = sumlev /(640 * 480);
}

//заполнение гистограммы
void Hist(HIST * h)
{
    U16 shx, cntx;
    U16 shy, cnty;
    //проверка и выставление границ
    if(((int)(h->startx) - _hist_size_x/2) < 0)
        h->startx += _hist_size_x/2;
    if(((int)(h->startx) + _hist_size_x/2) >= 640)
        h->startx -= _hist_size_x/2;
    if(((int)(h->starty) - _hist_size_y/2) < 0)
        h->starty += _hist_size_y/2;
    if(((int)(h->starty) + _hist_size_y/2) >= 480)
        h->starty -= _hist_size_x/2;
    shx = h->startx -_hist_size_x/2;
    shy = h->starty -_hist_size_y/2;
    //обнуление
    cntx = 0;
    while(cntx < _hist_size)
    {
        h->hist[cntx] = 0;
        cntx++;
    }
    //построение гистограммы
    cntx = 0;
    while(cntx < _hist_size_x) //значения координаты пикселя по X
    {
        cnty = 0;
        while(cnty < _hist_size_y) //значения координаты пикселя по Y
        {
            //увеличение значения ячейки для яркости пикселя wvbuf[shx + cntx][shy + cnty]
            h->hist[wvbuf[shx + cntx][shy + cnty]] += 1;
            cnty++;
        }
        cntx++;
    }
    //завершили построение
    h->work = false;
}

//сравнение гистограмм
int wCompareHist(HIST * h1, HIST * h2, HISTSEARCH * hs)
{
    int     cnt;
    float   val1, val2;
    float   err;
    //сравнение
    cnt     = 0;
    err     = 0.0f;
    while(cnt < _hist_size)
    {
        val1 = h1->hist[cnt];
        val2 = h2->hist[cnt];
        if((val1 + val2) > 0) //бывают и нулевые значения
            err += (float)(pow((double)(val1 - val2), 2) / (float)(val1 + val2));
        cnt++;
    }
    //ошибка
    hs->err = err;
    //
    return 0;
}

//поиск по гистограмме
//и фиксированной координате
int wSearchHist(void)
{
    U16     deltax;
    U16     deltay;
    float   radius;
    //поиск окончен ?!
    if(histsearch.serch_end)
        return 0;
    //сравнение гистограмм
    wCompareHist(&hist, &histregion, &histsearch);
    //значение ошибки меньше допустимой
    if(histsearch.err < histsearch.seterr)
    {
        //проверка региона
        deltax = abs(_centre_xregion - histsearch.stopx);
        deltay = abs(_centre_yregion - histsearch.stopy);
        radius = (float)(sqrt(pow((double)(deltax),2) + pow((double)(deltay),2)));
        if(radius < (float)_region_radius)
            histsearch.region = true;   //мы в регионе
        else
            histsearch.region = false;  //мы не в регионе
        //поиск окончен
        histsearch.serch_end = true;
        //фиксация новой гистограммы после поиска
        //hist.startx = histsearch.stopx;
        //hist.starty = histsearch.stopy;
        //Hist(&hist);
        return 0;
    }
    //
    if(histsearch.stopx < histsearch.x2)
        histsearch.stopx++;
    else
    {
        if(histsearch.stopy < histsearch.y3)
        {
            histsearch.stopx = histsearch.x1;
            histsearch.stopy++;
        }
    }
    //
    printf("startx=%i stopx=%i starty=%i stopy=%i x=%i y=%i search_end=%i\n",
           histsearch.x1, histsearch.x2, histsearch.y1, histsearch.y3, histsearch.stopx,histsearch.stopy,
           histsearch.serch_end);
    //histsearch.cntstep++;
    histregion.startx = histsearch.stopx;
    histregion.starty = histsearch.stopy;
    //
    return 0;
}
//

//рабочий поток
void * wWork_process(void * pr)
{
    strtxr  = _strtxr;
    strtyr  = _strtyr;
    szxr  = _szxr;
    szyr  = _szyr;
    stepr = _stepr;
    //
    while(WorkWork)
    {
        pthread_mutex_lock(&WorkMutex);
        //
        if(!(histsearch.serch_end))
            wRdCam(&wvbuf[0][0]);
        wMathExtr();
        if(hist.work)
            Hist(&hist);
        Hist(&histregion);
        wSearchHist();
        //
        pthread_mutex_unlock(&WorkMutex);
        usleep(_sleep_work);
    }
    pthread_exit(0);
}


//старт формирования шаблонной гистограммы
int wStartHist(U16 sx, U16 sy)
{
    pthread_mutex_lock(&WorkMutex);
    hist.startx = sx;
    hist.starty = sy;
    hist.work   = true;
    pthread_mutex_unlock(&WorkMutex);
    return 0;
}

//установка для поиска
int wsStartHist(U16 x, U16 y, float err)
{
    int res                 = 0;
    pthread_mutex_lock(&WorkMutex);
    //уствновка границ поиска
    //левый край зоны поиска
    if(x < _size_search / 2)
    {
        histsearch.x1 = 0;                                      //левый край в начале
        histsearch.x2 = _size_search;
        histsearch.x4 = histsearch.x1;                          //левый край в начале
        histsearch.x3 = histsearch.x2;
    }
    //правй край зоны поиска
    else if(x > (640 - (_size_search / 2)))
    {
        histsearch.x2 = 640;                                    //правый край в конце
        histsearch.x1 = histsearch.x2 - _size_search;
        histsearch.x4 = histsearch.x1;                          //правый край в конце
        histsearch.x3 = histsearch.x2;
    }
    else
    {
        histsearch.x1 = x - _size_search / 2;
        histsearch.x4 = x - _size_search / 2;
        histsearch.x2 = x + _size_search / 2;
        histsearch.x3 = x + _size_search / 2;
    }
    //верхний край
    if(y < _size_search / 2)
    {
        histsearch.y1 = 0;
        histsearch.y2 = histsearch.y1;
        histsearch.y3 = histsearch.y1 + _size_search;
        histsearch.y4 = histsearch.y3;
    }
    //нижний край
    else if(y > (480 - (_size_search / 2)))
    {
        histsearch.y3 = 480;
        histsearch.y4 = histsearch.y3;
        histsearch.y1 = histsearch.y3 -_size_search ;
        histsearch.y2 = histsearch.y1;
    }
    else
    {
        histsearch.y1 = y - _size_search / 2;
        histsearch.y2 = y - _size_search / 2;
        histsearch.y3 = y + _size_search / 2;
        histsearch.y4 = y + _size_search / 2;
    }
    histsearch.stopx = histsearch.x1;
    histsearch.stopy = histsearch.y1;
    //значение ошибки
    histsearch.seterr       = err;
    //поск не завершен
    histsearch.serch_end    = false;
    //сброс региона
    histsearch.region       = false;
    //
    pthread_mutex_unlock(&WorkMutex);
    return res;
}

//продолжение поиска
int wContinueSearch(void)
{
    int res = 0;
    pthread_mutex_lock(&WorkMutex);
    //начало поиска
    //histsearch.startx       = startx;
    //histsearch.starty       = starty;
    //
    //histsearch.stopx        = stopx;
    //histsearch.stopy        = stopy;
    //значение ошибки
    //histsearch.seterr       = err;
    //поск не завершен
    //histsearch.serch_end    = false;
    //сброс региона
    //histsearch.region       = false;
    //
    pthread_mutex_unlock(&WorkMutex);
    return res;
}
